@@ -1,11 +1,24 @@
 from rgsync.common import *
 from redisgears import getMyHashTag as hashtag
+from collections import OrderedDict
 
 class BaseSqlConnection():
     def __init__(self, user, passwd, db):
-        self.user = user
-        self.passwd = passwd
-        self.db = db
+        self._user = user
+        self._passwd = passwd
+        self._db = db
+
+    @property
+    def user(self):
+        return self._user() if callable(self._user) else self._user
+
+    @property
+    def passwd(self):
+        return self._passwd() if callable(self._passwd) else self._passwd
+
+    @property
+    def db(self):
+        return self._db() if callable(self._db) else self._db
 
     def _getConnectionStr(self):
         raise Exception('Can not use BaseSqlConnector _getConnectionStr directly')
@@ -14,7 +27,7 @@ class BaseSqlConnection():
         from sqlalchemy import create_engine
         ConnectionStr = self._getConnectionStr()
 
-        WriteBehindLog('Connect: connecting ConnectionStr=%s' % (ConnectionStr))
+        WriteBehindLog('Connect: connecting')
         engine = create_engine(ConnectionStr).execution_options(autocommit=True)
         conn = engine.connect()
         WriteBehindLog('Connect: Connected')
@@ -27,10 +40,21 @@ class MySqlConnection(BaseSqlConnection):
     def _getConnectionStr(self):
         return 'mysql+pymysql://{user}:{password}@{db}'.format(user=self.user, password=self.passwd, db=self.db)
 
+class PostgresConnection(BaseSqlConnection):
+    def __init__(self, user, passwd, db):
+        BaseSqlConnection.__init__(self, user, passwd, db)
+
+    def _getConnectionStr(self):
+        return 'postgresql://{user}:{password}@{db}'.format(user=self.user, password=self.passwd, db=self.db)
+
 class SQLiteConnection(BaseSqlConnection):
     def __init__(self, filePath):
         BaseSqlConnection.__init__(self, None, None, None)
-        self.filePath = filePath
+        self._filePath = filePath
+
+    @property
+    def filePath(self):
+        return self._filePath() if callable(self._filePath) else self._filePath
 
     def _getConnectionStr(self):
         return 'sqlite:////{filePath}?check_same_thread=False'.format(filePath=self.filePath)
@@ -42,10 +66,32 @@ class OracleSqlConnection(BaseSqlConnection):
     def _getConnectionStr(self):
         return 'oracle://{user}:{password}@{db}'.format(user=self.user, password=self.passwd, db=self.db)
 
+class MsSqlConnection(BaseSqlConnection):
+    def __init__(self, user, passwd, db, server, port, driver):
+        BaseSqlConnection.__init__(self, user, passwd, db)
+        self._server=server
+        self._port=port
+        self._driver=driver
+    @property
+    def server(self):
+        return self._server() if callable(self._server) else self._server
+    @property
+    def port(self):
+        return self._port() if callable(self._port) else self._port
+    @property
+    def driver(self):
+        return self._driver() if callable(self._driver) else self._driver
+    def _getConnectionStr(self):
+        return 'mssql+pyodbc://{user}:{password}@{server}:{port}/{db}?driver={driver}'.format(user=self.user, password=self.passwd, db=self.db, server=self.server, port=self.port, driver=self.driver)
+
 class SnowflakeSqlConnection(BaseSqlConnection):
     def __init__(self, user, passwd, db, account):
         BaseSqlConnection.__init__(self, user, passwd, db)
-        self.account = account
+        self._account = account
+
+    @property
+    def account(self):
+        return self._account() if callable(self._account) else self._account
 
     def _getConnectionStr(self):
         return 'snowflake://{user}:{password}@{account}/{db}'.format(user=self.user,
@@ -87,7 +133,7 @@ class BaseSqlConnector():
                 self.conn = self.connection.Connect()
                 if self.exactlyOnceTableName is not None:
                     shardId = 'shard-%s' % hashtag()
-                    result = self.conn.execute(self.sqlText('select val from %s where id=:id' % self.exactlyOnceTableName, {'id':shardId}))
+                    result = self.conn.execute(self.sqlText('select val from %s where id=:id' % self.exactlyOnceTableName), {'id':shardId})
                     res = result.first()
                     if res is not None:
                         self.exactlyOnceLastId = str(res['val'])
@@ -171,6 +217,39 @@ class MySqlConnector(BaseSqlConnector):
         if self.exactlyOnceTableName is not None:
             self.exactlyOnceQuery = GetUpdateQuery(self.exactlyOnceTableName, {'val', 'val'}, 'id')
 
+
+class PostgresConnector(MySqlConnector):
+
+    def __init__(self, connection, tableName, pk, exactlyOnceTableName=None):
+        BaseSqlConnector.__init__(self, connection, tableName, pk, exactlyOnceTableName)
+
+    def PrepereQueries(self, mappings):
+        def GetUpdateQuery(tableName, mappings, pk):
+
+            # OrderedDicts mean the same results each time
+            ordered_mappings = OrderedDict(mappings)
+            cols = ','.join(ordered_mappings.values())
+            values = [":{}".format(val) for kk, val in ordered_mappings.items() if not kk.startswith('_')]
+            values = list(values) + [":" + self.pk]
+
+            # prepare statement
+            value_stmt = ','.join(v for v in values)
+            update_stmts = ["{}=excluded.{}".format(v, v) for v in ordered_mappings.values() if not v.startswith('_')]
+            query = """INSERT INTO {} ({},{})
+            VALUES ({})
+            ON CONFLICT({}) DO UPDATE
+            SET
+            {}""".format(tableName, cols, self.pk,
+                    value_stmt,
+                    self.pk,
+                    ', '.join(update_stmts))
+            return query
+        self.addQuery = GetUpdateQuery(self.tableName, mappings, self.pk)
+        self.delQuery = 'delete from %s where %s=:%s' % (self.tableName, self.pk, self.pk)
+        if self.exactlyOnceTableName is not None:
+            self.exactlyOnceQuery = GetUpdateQuery(self.exactlyOnceTableName, {'val', 'val'}, 'id')
+
+
 class SQLiteConnector(MySqlConnector):
     def __init__(self, connection, tableName, pk, exactlyOnceTableName=None):
         MySqlConnector.__init__(self, connection, tableName, pk, exactlyOnceTableName)
@@ -193,6 +272,23 @@ class OracleSqlConnector(BaseSqlConnector):
         if self.exactlyOnceTableName is not None:
             self.exactlyOnceQuery = GetUpdateQuery(self.exactlyOnceTableName, 'id', ['id', 'val'], ['val'])
 
+class MsSqlConnector(BaseSqlConnector):
+    def __init__(self, connection, tableName, pk, exactlyOnceTableName=None):
+        BaseSqlConnector.__init__(self, connection, tableName, pk, exactlyOnceTableName)
+    def PrepereQueries(self, mappings):
+        values = [val for kk, val in mappings.items() if not kk.startswith('_')]
+        values_with_pkey = [self.pk] + values
+        def GetUpdateQuery(tableName, mappings, pk):
+            merge_into = "MERGE %s AS Target USING (VALUES (:%s)) AS Source (key1) ON (Target.%s = Source.key1)" % (tableName, pk, pk)
+            not_matched = "WHEN NOT MATCHED BY TARGET THEN INSERT (%s) VALUES (%s)" % (','.join(values_with_pkey), ','.join([':%s' % a for a in values_with_pkey]))
+            matched = "WHEN MATCHED THEN UPDATE SET %s" % (','.join(['Target.%s=:%s' % (a,a) for a in values]))
+            query = "%s %s %s;" % (merge_into, not_matched, matched)
+            return query
+        self.addQuery = GetUpdateQuery(self.tableName, mappings, self.pk)
+        self.delQuery = 'delete from %s where %s=:%s' % (self.tableName, self.pk, self.pk)
+        if self.exactlyOnceTableName is not None:
+            self.exactlyOnceQuery = GetUpdateQuery(self.exactlyOnceTableName, {'val', 'val'}, 'id')
+
 class SnowflakeSqlConnector(OracleSqlConnector):
     def __init__(self, connection, tableName, pk, exactlyOnceTableName=None):
         OracleSqlConnector.__init__(self, connection, tableName, pk, exactlyOnceTableName)
